/*
  Project: OscCVGateMod
  Description: Arduino 3+3ch CV/Gate-modulator
  Author: Staffan Melin, staffan.melin@oscillator.se
  License: GNU General Public License v3.0
  Version: 20200313
  Project site: http://www.oscillator.se/arduino

  v200101
  Initial version

  v20030913
  modulators for CV
  modulators for gate
  external sync

  V200406
  - cv mod: CVSequenceRandom
  - gate mod: GateSequence
  
  vFUTURE
  - CVSaw and derived classes: handle wavelength like 20 000?
  - CVSuperPot: three pots control one CV: speed, width, length of signal

*/

// if defined, open serial comm to enable error and status messages
// #define DEBUG


// libraries

// DAC
#include <Wire.h>
#include <Adafruit_MCP4725.h>



// constants

// buttons
#define PIN_POT_1 A2
#define PIN_POT_2 A1
#define PIN_POT_3 A0
#define PIN_POT_NONE 99

// The I2C Address. Download och run the I2C Scanner to look up the address
#define DAC_ADDRESS_1 0x62
#define DAC_ADDRESS_2 0x63

// connections

#define PIN_LED 13

// DAC select pins
#define PIN_DACA0_1 10
#define PIN_DACA0_2 8
#define PIN_DACA0_3 9

// GATEs
#define PIN_GATE_1 4
#define PIN_GATE_2 3
#define PIN_GATE_3 2

// DAC I2C communication
#define PIN_DAC_SDA A4
#define PIN_DAC_SCL A5

// value for max voltage from DAC
#define MAX_DAC 0x0FFF
// value for max voltage from ADC
#define MAX_ADC 0x0400

#define VOLTS_PER_NOTE 0.0833 // 1/12 V

// sequence
#define NOTE_ON 0
#define NOTE_OFF 1


// external clock
#define GATE_THRESHOLD 500 // ext gate threshold (of 0-1024) to trigger next note
#define PIN_CLOCK A7 // ext gate



// VARIABLES

// DAC
Adafruit_MCP4725 dac;

// note lengths
byte bpm = 120; // tempo, BPM (beats per minute)
byte t16 = 125; // length in ms of 1/16

// external input/clock/gate
unsigned int extLevel;
bool extClockBegin;
bool extClockSend;



// functions


// select DAC chip using DAC A0-pin (index from 1-)
void dacSelect(byte dacChip) {
  switch (dacChip) {
    case 1:
      digitalWrite(PIN_DACA0_1, LOW);
      digitalWrite(PIN_DACA0_2, HIGH);
      digitalWrite(PIN_DACA0_3, HIGH);
      break;
    case 2:
      digitalWrite(PIN_DACA0_1, HIGH);
      digitalWrite(PIN_DACA0_2, LOW);
      digitalWrite(PIN_DACA0_3, HIGH);
      break;
    case 3:
      digitalWrite(PIN_DACA0_1, HIGH);
      digitalWrite(PIN_DACA0_2, HIGH);
      digitalWrite(PIN_DACA0_3, LOW);
      break;
  }
}


// convert note number to DAC mV's
// DAC outputs from 0 to Vcc pin voltage (+5V)
// We have 1V/Oct, so 5 octaves
int midiToCV(byte pitch) {
  float voltsPerNote = 0.0833; // 1/12 V
  float mV; // from 0 to 0xFFF (4096);

  mV = 1000 * (pitch * voltsPerNote);

  return (constrain(map(mV, 0, 5000, 0, 4095), 0, 4095));
}


// as I mounted my pots the "wrong" way, I get max at clockwise left
// this function reads the pot and inverts it
int getPot(byte aPot) {
  int potValue = analogRead(aPot);
  return map(potValue, 0, MAX_ADC, MAX_ADC, 0);
}



/*
  CVModulator

  base class

  init arguments:
  aUseExtClock - use external clock [false/true]
  aValue - value to output [0 - MAX_DAC]
*/
class CVModulator {

  protected:

    uint16_t cvValue;
    bool cvUseExtClock;

  public:

    void init(bool aUseExtClock, uint16_t aValue) {
      cvUseExtClock = aUseExtClock;
      cvValue = aValue;
    }

    // if aUseExtClock == true and nextNote == true, change to a new/next value
    bool work(bool nextNote) {
      // return true if value has changed
      return (false);
    }

    uint16_t getCV() {
      return (cvValue);
    }

    void setCV(uint16_t aCV) {
      cvValue = aCV;
    }
};


/*
  CVFixed

  Outputs a fixed value

  init arguments:
  aUseExtClock - use external clock [false/true]
  aValue - value to output [0 - MAX_DAC]
*/
class CVFixed : public CVModulator {
  public:
    bool work(bool nextNote) {
      return (true);
    }
};


/*
  CVPot

  Outputs a value that can be controlled with a pot

  init arguments:
  aUseExtClock - use external clock [false/true]
  aFrom - lower boundary of pot [0 - MAX_DAC]
  aTo - upper boundary of pot [0 - MAX_DAC]
  aPot - which pot is in control
*/
class CVPot : public CVModulator {

  private:

    uint16_t cvFrom;
    uint16_t cvTo;
    byte cvPot;

  public:

    void init(bool aUseExtClock, uint16_t aFrom, uint16_t aTo, byte aPot) {
      cvUseExtClock = aUseExtClock; // IGNORE
      cvFrom = aFrom;
      cvTo = aTo;
      cvPot = aPot;
      this->work(false);
    }

    bool work(bool nextNote) {
      int potValue;
      potValue = getPot(cvPot);
      cvValue = map(potValue, 0, MAX_ADC, cvFrom, cvTo);
      return (true);
    }

};


/*
  CVRandom

  Outputs a random value whose width (variation) can optionally be controlled with a pot

  init arguments:
  aUseExtClock - use external clock [false/true]
  aFrom - lower boundary of value [0 - MAX_DAC]
  aTo - upper boundary of value [0 - MAX_DAC]
  aPeriod - length of signal in ms
  aPot - which pot is in control (set to PIN_POT_NONE to let class work without control)
*/
class CVRandom : public CVModulator {

  protected:

    unsigned long  lastTime;
    uint16_t cvFrom;
    uint16_t cvTo;
    uint16_t cvPeriod;
    uint16_t cvInterval;
    byte cvPot;

    void calcValue() {
      if (cvPot == PIN_POT_NONE) {
        cvValue = random(cvFrom, cvTo + 1);
      } else {
        float scale = float(getPot(cvPot)) / MAX_ADC;
        cvValue = random(1, cvInterval) * scale + (cvInterval - cvInterval * scale) / 2;
      }
    }

  public:

    void init(bool aUseExtClock, uint16_t aFrom, uint16_t aTo, uint16_t aPeriod, byte aPot) {
      cvUseExtClock = aUseExtClock;
      cvFrom = aFrom;
      cvTo = aTo;
      cvInterval = cvTo - cvFrom;
      cvPeriod = aPeriod;
      cvPot = aPot;
      calcValue();
      lastTime = millis();
    }

    bool work(bool nextNote) {
      unsigned long theTime = millis();
      bool cvChanged = false;

      if ((!cvUseExtClock && (theTime - lastTime >= cvPeriod)) || (cvUseExtClock && nextNote)) {
        calcValue();
        lastTime += cvPeriod;
        cvChanged = true;
      }
      
      return (cvChanged);
    }
};


/*
  CVRandomTime

  Outputs a random value at random intervals, the interval variation can optionally be controlled with a pot

  init arguments:
  aUseExtClock - use external clock [false/true]
  aFrom - lower boundary of value [0 - MAX_DAC]
  aTo - upper boundary of value [0 - MAX_DAC]
  aPot - which pot is in control (set to PIN_POT_NONE to let class work without control)
*/
class CVRandomTime : public CVModulator {

  protected:

    unsigned long lastTime;
    unsigned long interval;
    uint16_t cvFrom;
    uint16_t cvTo;
    byte cvPot;

    void calcValue() {
      cvValue = random(cvFrom, cvTo + 1);
      if (cvPot == PIN_POT_NONE) {
        interval = 5 * random(0, 1000); // wait from 0 - 2000 ms (20s)
      } else {
        interval = 5000 * (float(getPot(cvPot)) / MAX_ADC);
      }
    }

  public:

    void init(bool aUseExtClock, uint16_t aFrom, uint16_t aTo, byte aPot) {
      cvUseExtClock = aUseExtClock; // IGNORE
      cvFrom = aFrom;
      cvTo = aTo;
      cvPot = aPot;
      calcValue();
      lastTime = millis();

    }

    bool work(bool nextNote) {
      unsigned long theTime = millis();
      bool cvChanged = false;
      if (theTime - lastTime >= interval) {
        calcValue();
        lastTime = theTime;
        cvChanged = true;
      }
      return (cvChanged);
    }
};


/*
  CVRandomQuantized

  As CVRandom, but the output value is stepped to CV notes (1/12 V)
*/
class CVRandomQuantized : public CVRandom {
  protected:
    void calcValue() {
      cvValue = random(cvFrom, cvTo + 1) * VOLTS_PER_NOTE;

      if (cvPot != PIN_POT_NONE) {
        cvValue = MAX_DAC * float(getPot(cvPot) / MAX_ADC) + cvValue;
      }
    }
};


/*
  CVSequence

  Outputs a sequence of notes, the speed can optionally be controlled with a pot

  init arguments:
  aUseExtClock - use external clock [false/true]
  aPeriod - length of notes in ms
  aMidi - are notes in CV (false) och MIDI (true) format
  aNotes[] - array of note data, for CV [0 - MAX_DAC], for MIDI [0-60]
  aNumNotes - number of notes in aNotes[]
  aPot - which pot is in control (set to PIN_POT_NONE to let class work without control)
*/
class CVSequence : public CVModulator {

  protected:

    uint16_t cvPeriod;
    bool cvMidi; // are note values MIDI och CV
    uint16_t *cvNotes;
    byte cvNumNotes;
    byte cvPot;
    uint16_t cvPeriodScaled;

    unsigned long  lastTime;
    byte cvCurrentNote;

    void calcValue() {
      if (cvPot == PIN_POT_NONE) {
        cvPeriodScaled = cvPeriod;
      } else {
        float scale = float(getPot(cvPot)) / MAX_ADC;
        cvPeriodScaled = cvPeriod * scale * 4; // scale x4 so max pot is 4 x set period
      }
    }

  public:

    void init(bool aUseExtClock, uint16_t aPeriod, bool aMidi, uint16_t aNotes[], byte aNumNotes, byte aPot) {
      cvUseExtClock = aUseExtClock;
      cvPeriod = aPeriod;
      cvMidi = aMidi;
      cvNotes = aNotes; // pointer
      cvNumNotes = aNumNotes;
      cvPot = aPot;
      
      calcValue();
      lastTime = millis();
      cvCurrentNote = 0;
    }

    bool work(bool nextNote) {
      unsigned long theTime = millis();
      bool cvChanged = false;

      if ((!cvUseExtClock && (theTime - lastTime >= cvPeriodScaled)) || (cvUseExtClock && nextNote)) {
        cvCurrentNote++;
        if (cvCurrentNote >= cvNumNotes) {
          cvCurrentNote = 0;
        }
        if (cvMidi) {
          cvValue = midiToCV(cvNotes[cvCurrentNote]);
        } else {
          cvValue = cvNotes[cvCurrentNote];
        }
        lastTime = lastTime + cvPeriodScaled;
        calcValue();
        cvChanged = true;
      }
      
      return (cvChanged);
    }
};



/*
  CVSequenceRandom

  Outputs a random note from a sequence of notes, the speed can optionally be controlled with a pot.
  Pause between notes is also random (1-400% of period).

  init arguments:
  aUseExtClock - use external clock [false/true]
  aPeriod - length of notes in ms
  aMidi - are notes in CV (false) och MIDI (true) format
  aNotes[] - array of note data, for CV [0 - MAX_DAC], for MIDI [0-60]
  aNumNotes - number of notes in aNotes[]
  aPot - which pot is in control (set to PIN_POT_NONE to let class work without control)
*/
class CVSequenceRandom : public CVModulator {

  protected:

    uint16_t cvPeriod;
    bool cvMidi; // are note values MIDI och CV
    uint16_t *cvNotes;
    byte cvNumNotes;
    byte cvPot;
    uint16_t cvPeriodScaled;

    unsigned long  lastTime;
    byte cvCurrentNote;

    void calcValue() {
      if (cvPot == PIN_POT_NONE) {
        cvPeriodScaled = cvPeriod * (random(1, 400) / 100.0);
      } else {
        float scale = float(getPot(cvPot)) / MAX_ADC;
        cvPeriodScaled = cvPeriod * (random(1, 400) / 100.0) * scale * 4; // scale x4 so max pot is 4 x set period
      }
    }

  public:

    void init(bool aUseExtClock, uint16_t aPeriod, bool aMidi, uint16_t aNotes[], byte aNumNotes, byte aPot) {
      cvUseExtClock = aUseExtClock;
      cvPeriod = aPeriod;
      cvMidi = aMidi;
      cvNotes = aNotes; // pointer
      cvNumNotes = aNumNotes;
      cvPot = aPot;
      
      calcValue();
      lastTime = millis();
      cvCurrentNote = 0;
    }

    bool work(bool nextNote) {
      unsigned long theTime = millis();
      bool cvChanged = false;

      if ((!cvUseExtClock && (theTime - lastTime >= cvPeriodScaled)) || (cvUseExtClock && nextNote)) {
        cvCurrentNote  = random(0, cvNumNotes);
        if (cvMidi) {
          cvValue = midiToCV(cvNotes[cvCurrentNote]);
        } else {
          cvValue = cvNotes[cvCurrentNote];
        }
        lastTime = lastTime + cvPeriodScaled;
        calcValue();
        cvChanged = true;
      }
      
      return (cvChanged);
    }

    uint16_t getPeriod(void) {
      return (cvPeriodScaled);
    }
};




/*
  CVSaw

  Outputs a sawtooth wave (slow rise, instant fall), the speed can optionally be controlled with a pot

  init arguments:
  aUseExtClock - use external clock [false/true]
  aFrom - lower bound [0 - MAX_DAC]
  aTo - higher bound [0 - MAX_DAC]
  aWavelength - length of wave in ms
  aPot - which pot is in control (set to PIN_POT_NONE to let class work without control)
*/
class CVSaw : public CVModulator {

  protected:

    unsigned long  lastTime;
    unsigned long cvWavelength; // in ms
    uint16_t cvFrom;
    uint16_t cvTo;
    uint16_t cvStep;
    byte cvPot;
    uint16_t cvInterval;
    
    void calcValue() {
      if (cvPot == PIN_POT_NONE) {
        cvStep = cvInterval / cvWavelength; // change/ms
      } else {
        float scale = float(getPot(cvPot)) / MAX_ADC;
        cvStep = (cvInterval / (cvWavelength * scale));
      }
    }
    
  public:

    void init(bool aUseExtClock, uint16_t aFrom, uint16_t aTo, uint16_t aWavelength, byte aPot) {
      cvUseExtClock = aUseExtClock; // IGNORE
      cvFrom = aFrom;
      cvTo = aTo;
      cvWavelength = aWavelength;
      cvPot = aPot;
      cvValue = cvFrom;
      if (cvTo > cvFrom) {
        cvInterval = cvTo - cvFrom;
      } else {
        cvInterval = cvFrom - cvTo;
      }

      calcValue();
      lastTime = millis();
    }

    bool work(bool nextNote) {
      unsigned long theTime = millis();
      uint16_t nextStep;

      nextStep = (theTime - lastTime) * cvStep;
      
      if ((cvValue + nextStep) < cvTo) {
        cvValue += nextStep;
      } else {
        cvValue = cvFrom;
        calcValue();
      }
      lastTime = theTime;
      return (true);
    }
};



/*
  CVSawInv

  Outputs an inverted sawtooth wave (instant rise, slow fall), the speed can optionally be controlled with a pot

  init arguments:
  aUseExtClock - use external clock [false/true]
  aFrom - lower bound [0 - MAX_DAC]
  aTo - higher bound [0 - MAX_DAC]
  aWavelength - length of wave in ms
  aPot - which pot is in control (set to PIN_POT_NONE to let class work without control)
*/
class CVSawInv : public CVSaw {

  public:

    bool work(bool nextNote) {
      unsigned long theTime = millis();
      uint16_t nextStep;
      
      nextStep = (theTime - lastTime) * cvStep;
      // check so we don't get underflow
      if (cvValue > nextStep) {
        cvValue -= nextStep;
      } else {
        cvValue = cvFrom;
        calcValue();
      }
      lastTime = theTime;
      return (true);
    }
};


/*
  CVTriangle

  Outputs an inverted sawtooth wave (slow rise, slow fall), the speed can optionally be controlled with a pot

  init arguments:
  aUseExtClock - use external clock [false/true]
  aFrom - lower bound [0 - MAX_DAC]
  aTo - higher bound [0 - MAX_DAC]
  aWavelength - length of wave in ms
  aPot - which pot is in control (set to PIN_POT_NONE to let class work without control)
*/
class CVTriangle : public CVModulator {

  protected:
  
    unsigned long  lastTime;
    unsigned long cvWavelength; // in ms
    uint16_t cvFrom;
    uint16_t cvTo;
    byte cvPot;
    uint16_t cvStep;
    uint16_t cvInterval;
    bool cvRise;

    void calcValue() {
      if (cvPot == PIN_POT_NONE) {
        cvStep = cvInterval / cvWavelength; // change/ms
      } else {
        float scale = float(getPot(cvPot)) / MAX_ADC;
        cvStep = (cvInterval / (cvWavelength * scale));
      }
    }

  public:
  
    void init(bool aUseExtClock, uint16_t aFrom, uint16_t aTo, uint16_t aWavelength, byte aPot) {
      cvUseExtClock = aUseExtClock; // IGNORE
      cvFrom = aFrom;
      cvTo = aTo;
      cvWavelength = aWavelength;
      cvPot = aPot;

      cvInterval = cvTo - cvFrom;
      cvValue = cvFrom;
      calcValue();
      cvRise = true;

      lastTime = millis();
    }
    
    bool work(bool nextNote) {
      unsigned long theTime = millis();
      uint16_t nextStep;

      nextStep = (theTime - lastTime) * cvStep;
      
      if (cvRise) {
        if ((cvValue + nextStep) < cvTo) {
          cvValue += nextStep;
        } else {
          cvValue = cvTo;
          cvRise = false;
        }
      } else {
        if (cvValue > nextStep) {
          cvValue -= nextStep;
        } else {
          cvValue = cvFrom;
          calcValue();
          cvRise = true;
        }
      }
      lastTime = theTime;
      return (true);
    }
};


/*
  CVSquare

  Outputs a square wave (instant rise, instant fall), the speed can optionally be controlled with a pot

  init arguments:
  aUseExtClock - use external clock [false/true]
  aFrom - lower bound [0 - MAX_DAC]
  aTo - higher bound [0 - MAX_DAC]
  aWavelength - length of wave in ms
  aPot - which pot is in control (set to PIN_POT_NONE to let class work without control)
*/
class CVSquare : public CVModulator {

  protected:

    unsigned long  lastTime;
    unsigned long cvWavelength; // in ms
    uint16_t cvFrom;
    uint16_t cvTo;
    byte cvPot;
    bool cvHigh;
    uint16_t cvLength;

    void calcValue() {
      if (cvPot == PIN_POT_NONE) {
        cvLength = cvWavelength;
      } else {
        float scale = float(getPot(cvPot)) / MAX_ADC;
        cvLength = cvWavelength * scale;
      }
    }
    
  public:

    void init(bool aUseExtClock, uint16_t aFrom, uint16_t aTo, uint16_t aWavelength, byte aPot) {
      cvUseExtClock = aUseExtClock; // IGNORE
      
      cvFrom = aFrom;
      cvTo = aTo;
      cvWavelength = aWavelength; // length of half-cycle
      cvPot = aPot;

      cvValue = cvFrom;
      calcValue();
      lastTime = millis() - cvWavelength + 1; // make sure we get a change on first iteration
      cvHigh = false;
    }

    bool work(bool nextNote) {
      unsigned long theTime = millis();
      bool bChanged = false;
      
      if (theTime - lastTime > cvLength) {
        calcValue();
        cvHigh = !cvHigh;
        lastTime = theTime;
        bChanged = true;
      }

      if (cvHigh) {
        cvValue = cvTo;
      } else {
        cvValue = cvFrom;
      }

      return (bChanged);
    }
};


/*
  CVSine

  Outputs a sine wave, the speed can optionally be controlled with a pot

  init arguments:
  aUseExtClock - use external clock [false/true]
  aFrom - lower bound [0 - MAX_DAC]
  aTo - higher bound [0 - MAX_DAC]
  aWavelength - length of wave in ms
  aPot - which pot is in control (set to PIN_POT_NONE to let class work without control)
*/
class CVSine : public CVModulator {

  protected:

    unsigned long  lastTime;
    unsigned long cvWavelength; // in ms
    uint16_t cvFrom;
    uint16_t cvTo;
    byte cvPot;
    uint16_t cvStep;
    uint16_t cvCurrentPos;
    uint16_t cvHeight;
    uint16_t cvLength;

    void calcValue() {
      if (cvPot == PIN_POT_NONE) {
        cvLength = cvWavelength;
      } else {
        float scale = float(getPot(cvPot)) / MAX_ADC;
        cvLength = cvWavelength * scale;
      }
    }

  public:

    void init(bool aUseExtClock, uint16_t aFrom, uint16_t aTo, uint16_t aWavelength, byte aPot) {
      cvUseExtClock = aUseExtClock; // IGNORE
      cvFrom = aFrom;
      cvTo = aTo;
      cvWavelength = aWavelength;
      cvPot = aPot;

      cvValue = cvFrom;
      cvCurrentPos = 0;
      cvHeight = (cvTo - cvFrom) / 2;
      calcValue();
      lastTime = millis();
    }

    bool work(bool nextNote) {
      bool bChanged = false;
      unsigned long theTime = millis();
      uint16_t timePassed;
      float fraction;

      // calc new value each 10 ms
      timePassed = (theTime - lastTime);
      if (timePassed > 10) {
        cvCurrentPos += timePassed;
        if (cvCurrentPos >= cvLength) {
          cvCurrentPos = 0;
        }
        fraction = float(cvCurrentPos) / cvLength; 
        cvValue = sin(fraction * 2 * 3.14) * cvHeight + cvHeight;

        calcValue();
        bChanged = true;
        lastTime = theTime;
      }

      return (bChanged);
    }
};


// CVWanderer
// walking random CV
  // cv1.init(false, 0, MAX_DAC, 1, 10, PIN_POT_1, PIN_POT_2); // CVWanderer

/*
  CVWanderer

  Outputs a signal that wanders about randomally, the amount of change and speed can optionally be controlled with a pot

  init arguments:
  aUseExtClock - use external clock [false/true]
  aFrom - lower bound [0 - MAX_DAC]
  aTo - higher bound [0 - MAX_DAC]
  aSpeed - speed of wandering, try values from 1-100
  aChance - probability of wanderer changing direction [1-99]
  aPotWidth - which pot is in control of amount of change (set to PIN_POT_NONE to let class work without control)
  aPotSpeed - which pot is in control of speed (set to PIN_POT_NONE to let class work without control)
*/
class CVWanderer : public CVModulator {

  protected:
  
    unsigned long  lastTime;
    unsigned long cvWavelength; // in ms
    uint16_t cvFrom;
    uint16_t cvTo;
    uint16_t cvSpeed; // how many ms must pass before 1 step
    byte cvChance; // chance of changing direction (each ms)
    byte cvPotWidth;
    byte cvPotSpeed;
    uint16_t cvStep;
    uint16_t cvSpeedScaled;
    bool cvRise;

    void calcValue() {
      if (cvPotWidth == PIN_POT_NONE) {
        cvStep = 1;
      } else {
        float scale = float(getPot(cvPotWidth)) / MAX_ADC;
        cvStep = scale * 100;
      }
      if (cvPotSpeed == PIN_POT_NONE) {
        cvSpeedScaled = cvSpeed;
      } else {
        float scale = float(getPot(cvPotSpeed)) / MAX_ADC;
        cvSpeedScaled = cvSpeed * (100 * (1-scale));
      }
      if (random(1, 100) < cvChance) {
        cvRise = !cvRise;
      }
      
    }

  public:
  
    void init(bool aUseExtClock, uint16_t aFrom, uint16_t aTo, uint16_t aSpeed, byte aChance, byte aPotWidth, byte aPotSpeed) {
      cvUseExtClock = aUseExtClock; // IGNORE
      cvFrom = aFrom;
      cvTo = aTo;
      cvSpeed = aSpeed;
      cvChance = aChance;
      cvPotWidth = aPotWidth;
      cvPotSpeed = aPotSpeed;

      cvValue = (cvTo - cvFrom) / 2;
      cvRise = true;
      calcValue();

      lastTime = millis();
    }
    
    bool work(bool nextNote) {
      unsigned long theTime = millis();

      calcValue();

      if (theTime - lastTime > cvSpeedScaled) {
      
        if (cvRise) {
          if ((cvValue + cvStep) < cvTo) {
            cvValue += cvStep;
          } else {
            cvValue = cvTo;
            cvRise = false;
          }
        } else {
          if (cvValue > cvStep) {
            cvValue -= cvStep;
          } else {
            cvValue = cvFrom;
            cvRise = true;
          }
        }
        lastTime = theTime;
      }
      return (true);
    }
};



/*
  GateModulator

  base class

  init arguments:
  aUseExtClock - use external clock [false/true]
  aValue - value to output [false/true]
*/
class GateModulator {

  protected:

    bool gateValue;
    bool gateUseExtClock;

  public:

    void init(bool aUseExtClock, bool aValue) {
      gateUseExtClock = aUseExtClock;
      gateValue = aValue;
    }

    // if aUseExtClock == true and nextNote == true, change to a new/next value
    bool work(bool nextNote) {
      // return true if value has changed
      return (false);
    }

    bool getGate() {
      return (gateValue);
    }

    void setGate(bool aGate) {
      gateValue = aGate;
    }

    void setPeriod(unsigned long aPeriod) {
      //gatePeriodScaled = aPeriod;
    }
};


/*
  GateFixed

  Outputs a fixed value

  init arguments:
  aUseExtClock - use external clock [false/true]
  aValue - value to output [false/true]
*/
class GateFixed : public GateModulator {
  public:

    void init(bool aUseExtClock, bool aValue) {
      gateUseExtClock = aUseExtClock; // IGNORE
      gateValue = aValue;
    }

    bool work(bool nextNote) {
      return (true);
    }
};


/*
  GateClock

  Outputs a clock pulse, the period can optionally be controlled by a pot

  init arguments:
  aUseExtClock - use external clock [false/true]
  aPeriod - 
  aLength - length (as percent) of period 
  aPotPeriod
*/
class GateClock : public GateModulator {

  protected:

    unsigned long lastTime;
    unsigned long lastGateTime;
    unsigned long gatePeriod;
    int gateLength; // percentage 0-100
    byte gatePotPeriod;
    unsigned long gateLengthTime;
    unsigned long gatePeriodScaled;
    unsigned long gateLengthTimeScaled;

    void calcValue() {
      if (gatePotPeriod == PIN_POT_NONE) {
        gatePeriodScaled = gatePeriod;
        gateLengthTimeScaled = gateLengthTime;
      } else {
        float scale = float(getPot(gatePotPeriod)) / MAX_ADC;
        gatePeriodScaled = gatePeriod * scale;
        gateLengthTimeScaled = gatePeriodScaled * (float(gateLength) / 100);
      }
    }

  public:

    void init(bool aUseExtClock, unsigned long aPeriod, int aLength, byte aPotPeriod) {
      gateUseExtClock = aUseExtClock; // IGNORE
      gatePeriod = aPeriod;
      gateLength = aLength;
      gateLengthTime = aPeriod * (float(aLength) / 100);
      gatePotPeriod = aPotPeriod;

      calcValue();
      lastTime = millis();
      gateValue = true;
    }

    bool work(bool nextNote) {
      unsigned long theTime = millis();

      if (gateValue) {
        if (theTime - lastTime >= gateLengthTimeScaled) {
          gateValue = false;
        }
      } else {
        if (theTime - lastTime >= gatePeriodScaled) {
          lastTime += gatePeriodScaled;
          gateValue = true;
          calcValue();
        }

      }
      return (true);
    }

    void setPeriod(unsigned long aPeriod) {
      gatePeriodScaled = aPeriod;
    }

};


// GateRandom
// outputs clock with random tempo controlled by pot 
class GateRandom : public GateModulator {

  protected:

    unsigned long lastTime;
    byte gateTempoPot;

    unsigned long gatePeriod;
    int gateLength; // percentage 0-100
    unsigned long gateLengthTimeScaled;
    unsigned long gatePeriodScaled;

    void calcValue() {
      float scale;
      
      if (gateTempoPot == PIN_POT_NONE) {
        scale = float(random(1, 100)) /10 ;
      } else {
        float scalePot = float(getPot(gateTempoPot)) / MAX_ADC;
        scale = float(random(1, 100)) * scalePot / 10; 
      }
      gatePeriodScaled = gatePeriod * scale;
      gateLengthTimeScaled = gatePeriodScaled * (float(gateLength) / 100);
    }

  public:

    void init(bool aUseExtClock, unsigned long aPeriod, int aLength, byte aPot) {
      gateUseExtClock = aUseExtClock; // IGNORE
      gatePeriod = aPeriod;
      gateLength = aLength;
      gateTempoPot = aPot;
      calcValue();
      lastTime = millis();
      gateValue = true;
    }

    bool work(bool nextNote) {
      unsigned long theTime = millis();

      if (gateValue) {
        if (theTime - lastTime >= gateLengthTimeScaled) {
          gateValue = false;
        }
      } else {
        if (theTime - lastTime >= gatePeriodScaled) {
          lastTime += gatePeriodScaled;
          gateValue = true;
          calcValue();
        }
      }
      return (true);
      //return (gateValuePrev != gateValue);
    }
};


/*
  GateSequence

  Outputs a sequence of notes, the speed can optionally be controlled with a pot

  init arguments:
  aUseExtClock - use external clock [false/true]
  aPeriod - length of notes in ms
  aNotes[] - array of note data, for CV [0 - MAX_DAC], for MIDI [0-60]
  aNumNotes - number of notes in aNotes[]
  aPot - which pot is in control (set to PIN_POT_NONE to let class work without control)
*/
class GateSequence : public GateModulator {

  protected:

    uint16_t gatePeriod;
    uint16_t *gateNotes; // 16 bits are not necessary
    byte gateNumNotes;
    byte gatePot;
    uint16_t gatePeriodScaled;

    unsigned long lastTime;
    byte gateCurrentNote;

    void calcValue() {
      if (gatePot == PIN_POT_NONE) {
        gatePeriodScaled = gatePeriod;
      } else {
        float scale = float(getPot(gatePot)) / MAX_ADC;
        gatePeriodScaled = gatePeriod * scale * 4; // scale x4 so max pot is 4 x set period
      }
    }

  public:

    void init(bool aUseExtClock, uint16_t aPeriod, uint16_t aNotes[], byte aNumNotes, byte aPot) {
      gateUseExtClock = aUseExtClock;
      gatePeriod = aPeriod;
      gateNotes = aNotes; // pointer
      gateNumNotes = aNumNotes;
      gatePot = aPot;
      
      calcValue();
      lastTime = millis();
      gateCurrentNote = 0;
    }

    bool work(bool nextNote) {
      unsigned long theTime = millis();
      bool gateChanged = false;

      if ((!gateUseExtClock && (theTime - lastTime >= gatePeriodScaled)) || (gateUseExtClock && nextNote)) {
        gateCurrentNote++;
        if (gateCurrentNote >= gateNumNotes) {
          gateCurrentNote = 0;
        }
        gateValue = (gateNotes[gateCurrentNote] == NOTE_ON);
        lastTime = lastTime + gatePeriodScaled;
        calcValue();
        gateChanged = true;
      }
      
      return (gateChanged);
    }
};



// CVFixed cv1;
// CVPot cv1;
// CVRandom cv1;
// CVRandomTime cv1;
// CVRandomQuantized cv1;
// CVSequence cv1;
// CVSaw cv1;
// CVSawInv cv1;
// CVTriangle cv1;
// CVSquare cv1;
// CVSine cv1;
// CVWanderer cv1;

CVSequenceRandom cv1;
CVTriangle cv2;
CVWanderer cv3;

GateClock gate1;
//GateFixed gate1;
GateSequence gate2;
GateRandom gate3;



////////////////////////////////////////////////////



// SETUP

void setup() {

  //pinMode(PIN_LED, OUTPUT);

  // DAC select
  pinMode(PIN_DACA0_1, OUTPUT);
  pinMode(PIN_DACA0_2, OUTPUT);
  pinMode(PIN_DACA0_3, OUTPUT);

  // DAC
  pinMode(PIN_DAC_SDA, OUTPUT);
  pinMode(PIN_DAC_SCL, OUTPUT);

  // GATEs
  pinMode(PIN_GATE_1, OUTPUT);
  pinMode(PIN_GATE_2, OUTPUT);
  pinMode(PIN_GATE_3, OUTPUT);

  // EXT CLOCK
  pinMode(PIN_CLOCK, INPUT);

  // POTs
  pinMode(PIN_POT_1, INPUT);
  pinMode(PIN_POT_2, INPUT);
  pinMode(PIN_POT_3, INPUT);

  // init DACs
  dac.begin(DAC_ADDRESS_1);
  dacSelect(1);

  // set all gates on
  digitalWrite(PIN_GATE_1, HIGH);
  digitalWrite(PIN_GATE_2, HIGH);
  digitalWrite(PIN_GATE_3, HIGH);

  // cv1.init(false, 2500); // CVFixed
  // cv1.init(false, 1500, 2000, PIN_POT_1); // CVPot
  // cv1.init(false, 0, MAX_DAC, t16*4, PIN_POT_NONE); // CVRandom
  // cv1.init(false, 0, MAX_DAC, 2000, PIN_POT_1); // CVSaw
  // cv1.init(false, MAX_DAC, 0, 2000, PIN_POT_1); // CVSawInv
  // cv1.init(false, 0, MAX_DAC, 2000, PIN_POT_1); // CVTriangle
  // cv1.init(false, 0, MAX_DAC, 2000, PIN_POT_1); // CVSquare
  // cv1.init(false, 0, MAX_DAC, 2000, PIN_POT_1); // CVSine
  // cv1.init(false, 0, MAX_DAC, 1, 10, PIN_POT_1, PIN_POT_2); // CVWanderer

  // notes must be less than 60 (if midi)
  // uint16_t notes[8] = {20, 20, 21, 20, 21, 23, 25, 20};
  // uint16_t notes[8] = {1800, 1900, 2000, 2100, 2200, 1800, 1900, 2000};
  // cv1.init(true, t16*2, true, notes, 8, PIN_POT_1); // CVSequence (MIDI)

  uint16_t notes[8] = {20, 20, 21, 20, 21, 23, 25, 20};
  cv1.init(false, t16*4, true, notes, 8, PIN_POT_1); // CVSequenceRandom
  cv2.init(false, 0, MAX_DAC, 2000, PIN_POT_2); // CVTriangle
  cv3.init(false, 0, MAX_DAC, 1, 10, PIN_POT_NONE, PIN_POT_NONE); // CVWanderer

  // gate1.init(false, true); // GateFixed
  // gate1.init(false, t16*2, 50, PIN_POT_2, PIN_POT_3); // GateClock
  // gate1.init(false, t16*2, 20, PIN_POT_3); // GateRandom

  gate1.init(false, t16*4, 50, PIN_POT_NONE); // GateClock

  uint16_t gates[16] = {NOTE_ON, NOTE_OFF, NOTE_ON, NOTE_OFF, NOTE_OFF, NOTE_OFF, NOTE_OFF, NOTE_OFF, NOTE_OFF, NOTE_OFF, NOTE_OFF, NOTE_OFF, NOTE_OFF, NOTE_OFF, NOTE_OFF, NOTE_OFF};
  gate2.init(false, t16*1, gates, 16, PIN_POT_2); // GateSequence

  //gate3.init(false, t16*4, 50, PIN_POT_3); // GateClock
  gate3.init(false, t16*2, 50, PIN_POT_3); // GateRandom

  extClockBegin = false;
  extClockSend = false;

#ifdef DEBUG
  Serial.begin(115200);
#endif

  randomSeed(analogRead(0));
}



// LOOP

void loop() {

  // external clock

  extLevel = analogRead(PIN_CLOCK);

  if (extLevel > GATE_THRESHOLD) {

    if (!extClockBegin) {
      extClockSend = true;
    } else {
      extClockSend = false;
    }
    extClockBegin = true;
  } else {
    extClockBegin = false;
    extClockSend = false;
  }

  // cv: process and send

  dacSelect(1);
  if (cv1.work(extClockSend)) {
    dac.setVoltage(cv1.getCV(), false);

    // sync CV and Gate; only works for CVSequenceRandom() and GateClock
    gate1.setPeriod(cv1.getPeriod());
  }

  dacSelect(2);
  if (cv2.work(extClockSend)) {
    dac.setVoltage(cv2.getCV(), false);
  }

  dacSelect(3);
  if (cv3.work(extClockSend)) {
    dac.setVoltage(cv3.getCV(), false);
  }

  // gate: process and send

  if (gate1.work(extClockSend)) {
    if (gate1.getGate()) {
      digitalWrite(PIN_GATE_1, HIGH);
    } else {
      digitalWrite(PIN_GATE_1, LOW);
    }
  }

  if (gate2.work(extClockSend)) {
    if (gate2.getGate()) {
      digitalWrite(PIN_GATE_2, HIGH);
    } else {
      digitalWrite(PIN_GATE_2, LOW);
    }
  }

  if (gate3.work(extClockSend)) {
    if (gate3.getGate()) {
      digitalWrite(PIN_GATE_3, HIGH);
    } else {
      digitalWrite(PIN_GATE_3, LOW);
    }
  }

  delay(1);

}
